GPU_IDS=$1
PYTHON_INTERPRETER=$2
SCRIPT_PATH=$3
CONFIG_PATH=$4
# echo "CUDA_VISIBLE_DEVICES=$GPU_IDS $PYTHON_INTERPRETER $SCRIPT_PATH $CONFIG_PATH"
CUDA_VISIBLE_DEVICES=$GPU_IDS $PYTHON_INTERPRETER $SCRIPT_PATH $CONFIG_PATH
